#!/usr/bin/env python
'''
booxter2csv.py

Takes XML output from Booxter (by Deep Prose) v1 or 2
and outputs CSV files suitable for import into Open Office,
GCStar, Goodreads etc.

Requirements:  Python 2.5 or newer
'''
__author__  = 'furicle brian.mckee@gmail.com'
__version__ = '0.5'

# import required modules
import sys
from optparse import OptionParser
import xml.etree.ElementTree as ET
import xml.parsers.expat
import csv

# Parse arguments received
usage = 'Usage: %prog [options] XMLFILE'
parser = OptionParser(usage, version='%prog '+__version__)
parser.add_option('-d', '--multi-item-delimiter', dest='multiItemDelimiter', default='|',
                  help='Multiple item (e.g. co-authors) record delimiter Default is %default')
parser.add_option('-g', '--goodreads', action='store_true', dest='goodreads', default=False,
                  help='output CSV file suitable for import in Goodreads.com. Default is %default')
parser.add_option('-o', '--output-file', dest='outputFile', default='books.csv',
                  help='CSV output FILENAME Default is %default')
parser.add_option('-s', '--strip', action='store_true', dest='strip', default=False,
                  help='remove line feeds and carriage returns from inside records. Default is %default')
parser.add_option('-v', '--verbose', action='store_true', dest='verbose', default=False,
                  help='running commentary about progress and options provided. Default is %default')
(options, args) = parser.parse_args()

csvFileOut = options.outputFile

# make sure an input file specified
if len(args) == 1 :
    xmlFileIn  = args[0]
else:
    parser.error('No XMLFILE in provided')

# fields with multiple values are treated as sub elements in the xml file
# e.g. a book with two authors
# they need to be combined into one field separated by a delimiter that's
# distinct from those already in use by CSV
invalidDelimiters = ('"',"'","\n")
if options.multiItemDelimiter in invalidDelimiters:
    parser.error('Internal record delimiter can\'t be a invalid in a CSV file')

# verbose is actually closer to debug than verbose I suppose...
if options.verbose:
    print
    print 'Welcome to booxml2csv'
    print
    if options.goodreads: print '-- Goodreads Fields'
    if options.strip    : print '-- Strip Carriage Returns From Output'
    print 'Input File  %s' % xmlFileIn
    print 'Output File %s' % csvFileOut
    print

# open and parse xml file and set books as the base element
if options.verbose: print 'Loading %s input file...' % xmlFileIn
try:
    library=ET.parse(xmlFileIn)
    books = library.getroot()
except IOError, (errno, strerror):
    print "I/O error(%s): %s opening %s" % (errno, strerror, xmlFileIn)
    sys.exit(errno)
except xml.parsers.expat.ExpatError, e:
    print 'Parse Error: %s' % e
    print 'Is it valid XML?'
    sys.exit(1)
if options.verbose: print  '%s records found' % len(books)

# average rating, original published date, read, not booxter fields, but here so they get blank filled
goodReadsXref=[['title','Title'], ['authors','Author'], ['isbn','ISBN'], ['rating','My Rating'],
        ['averageRating','Average Rating'], ['publisher','Publisher'], ['format','Binding'],
        ['datePublished','Year Published'], ['originalPublished','Original Publication Year'],
        ['dateRead','Date Read'], ['dateAdded','Date Added'], ['Genre','Bookshelves'],
        ['Notes','My Review']]

multiItemFields = {'authors':'author', 'locSubjects':'locSubject'}

# get a list of all the fields present in all books
fields = []
for book in books:
    for field in book:
        fields.append(field.tag)

# create a unique list of fields
uniqueFields = []
for field in fields:
    if field not in uniqueFields:
        uniqueFields.append(field)

# and sort them alpha
uniqueFields.sort()
if options.verbose: print  '%s unique fields found' % len(uniqueFields)

# open csv out file
if options.verbose: print  'Opening csv file %s...' % csvFileOut
try:
    writer=csv.writer(open(csvFileOut,'wb'))
except IOError, (errno, strerror):
    print "I/O error(%s): %s opening %s" % (errno, strerror, csvFileOut)
    sys.exit(errno)

# print header row to csv file
headerFields = []
if options.goodreads:
    for field in goodReadsXref:
        headerFields.append(field[1])
else:
    headerFields = uniqueFields

if options.verbose: print  'Writing headers to %s...' % csvFileOut
writer.writerow(headerFields)

if options.verbose: print  'Writing rows to %s...' % csvFileOut
if options.goodreads:
    dataFields = goodReadsXref
else:
    dataFields = uniqueFields

for book in books:
# build row
    row=[]
    for item in dataFields:
        if options.goodreads:
            field = item[0]
        else:
            field = item
        cell=u''
# check if this is a multiple value field and handle as such
        if field in multiItemFields:
            searchPattern = field+'/'+multiItemFields[field]
            subfields = book.findall(searchPattern)
            subCells = []
            for subfield in subfields:
                subCells.append(subfield.text)
            cell=options.multiItemDelimiter.join(subCells)
# regular single value field
        else:
            cell = book.findtext(field)
# if that field doesn't exist it returns None, which isn't what we want in the csv output
        if cell is None:
            cell = ''
# the csv module can't handle unicode input, but utf-8 is ok
        cell = cell.encode('utf-8')
# remove line feeds if requested
        if options.strip:
            cell=cell.replace('\n', ' ')
            cell=cell.strip()
        row.append(cell)
# write out the row to the csv file
    writer.writerow(row)

if options.verbose: print  '\nAll Done!\n',
